/*
 * Copyright 2012-2013 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __DBUS_DATA_H__
#define __DBUS_DATA_H__ 1

#define INDICATOR_MESSAGES_DBUS_NAME               "org.ayatana.indicator.messages"
#define INDICATOR_MESSAGES_DBUS_OBJECT             "/org/ayatana/indicator/messages"

#define INDICATOR_MESSAGES_DBUS_SERVICE_OBJECT     "/org/ayatana/indicator/messages/service"
#define INDICATOR_MESSAGES_DBUS_SERVICE_INTERFACE  "org.ayatana.indicator.messages.service"

#endif /* __DBUS_DATA_H__ */
